@@ -19,9 +19,11 @@ License along with this library.  If not, see <http://www.gnu.org/licenses/>.
 *********************************************************************/
 #include "display.h"
 #include "compositor_interface.h"
+#include "datadevicemanager_interface.h"
 #include "output_interface.h"
 #include "seat_interface.h"
 #include "shell_interface.h"
+#include "subcompositor_interface.h"
 
 #include <QCoreApplication>
 #include <QDebug>
@@ -41,12 +43,14 @@ public:
     Private(Display *q);
     void flush();
     void setRunning(bool running);
+    void installSocketNotifier();
 
     wl_display *display = nullptr;
     wl_event_loop *loop = nullptr;
     QString socketName = QStringLiteral("wayland-0");
     bool running = false;
     QList<OutputInterface*> outputs;
+    QVector<ClientConnection*> clients;
 
 private:
     Display *q;
@@ -57,11 +61,26 @@ Display::Private::Private(Display *q)
 {
 }
 
+void Display::Private::installSocketNotifier()
+{
+    if (!QCoreApplication::instance()) {
+        return;
+    }
+    int fd = wl_event_loop_get_fd(loop);
+    if (fd == -1) {
+        qWarning() << "Did not get the file descriptor for the event loop";
+        return;
+    }
+    QSocketNotifier *m_notifier = new QSocketNotifier(fd, QSocketNotifier::Read, q);
+    QObject::connect(m_notifier, &QSocketNotifier::activated, q, [this] { flush(); } );
+    QObject::connect(QCoreApplication::eventDispatcher(), &QAbstractEventDispatcher::aboutToBlock, q, [this] { flush(); });
+    setRunning(true);
+}
+
 Display::Display(QObject *parent)
     : QObject(parent)
     , d(new Private(this))
 {
-    connect(QCoreApplication::eventDispatcher(), &QAbstractEventDispatcher::aboutToBlock, this, [this] { d->flush(); });
 }
 
 Display::~Display()
@@ -94,24 +113,34 @@ QString Display::socketName() const
     return d->socketName;
 }
 
-void Display::start()
+void Display::start(StartMode mode)
 {
     Q_ASSERT(!d->running);
     Q_ASSERT(!d->display);
     d->display = wl_display_create();
-    if (wl_display_add_socket(d->display, qPrintable(d->socketName)) != 0) {
-        return;
+    if (mode == StartMode::ConnectToSocket) {
+        if (wl_display_add_socket(d->display, qPrintable(d->socketName)) != 0) {
+            return;
+        }
     }
 
     d->loop = wl_display_get_event_loop(d->display);
-    int fd = wl_event_loop_get_fd(d->loop);
-    if (fd == -1) {
-        qWarning() << "Did not get the file descriptor for the event loop";
-        return;
-    }
-    QSocketNotifier *m_notifier = new QSocketNotifier(fd, QSocketNotifier::Read, this);
-    connect(m_notifier, &QSocketNotifier::activated, this, [this] { d->flush(); } );
-    d->setRunning(true);
+    d->installSocketNotifier();
+}
+
+void Display::startLoop()
+{
+    Q_ASSERT(!d->running);
+    Q_ASSERT(d->display);
+    d->installSocketNotifier();
+}
+
+void Display::dispatchEvents(int msecTimeout)
+{
+    Q_ASSERT(!d->running);
+    Q_ASSERT(d->display);
+    wl_event_loop_dispatch(d->loop, msecTimeout);
+    wl_display_flush_clients(d->display);
 }
 
 void Display::terminate()
@@ -164,6 +193,20 @@ SeatInterface *Display::createSeat(QObject *parent)
     return seat;
 }
 
+SubCompositorInterface *Display::createSubCompositor(QObject *parent)
+{
+    auto c = new SubCompositorInterface(this, parent);
+    connect(this, &Display::aboutToTerminate, c, [this,c] { delete c; });
+    return c;
+}
+
+DataDeviceManagerInterface *Display::createDataDeviceManager(QObject *parent)
+{
+    auto m = new DataDeviceManagerInterface(this, parent);
+    connect(this, &Display::aboutToTerminate, m, [this,m] { delete m; });
+    return m;
+}
+
 void Display::createShm()
 {
     Q_ASSERT(d->running);
@@ -206,6 +249,48 @@ QList< OutputInterface* > Display::outputs() const
     return d->outputs;
 }
 
+ClientConnection *Display::getConnection(wl_client *client)
+{
+    Q_ASSERT(client);
+    auto it = std::find_if(d->clients.constBegin(), d->clients.constEnd(),
+        [client](ClientConnection *c) {
+            return c->client() == client;
+        }
+    );
+    if (it != d->clients.constEnd()) {
+        return *it;
+    }
+    // no ConnectionData yet, create it
+    auto c = new ClientConnection(client, this);
+    d->clients << c;
+    connect(c, &ClientConnection::disconnected, this,
+        [this] (ClientConnection *c) {
+            const int index = d->clients.indexOf(c);
+            Q_ASSERT(index != -1);
+            d->clients.remove(index);
+            Q_ASSERT(d->clients.indexOf(c) == -1);
+            emit clientDisconnected(c);
+        }
+    );
+    emit clientConnected(c);
+    return c;
+}
+
+QVector< ClientConnection* > Display::connections() const
+{
+    return d->clients;
+}
+
+ClientConnection *Display::createClient(int fd)
+{
+    Q_ASSERT(fd != -1);
+    Q_ASSERT(d->display);
+    wl_client *c = wl_client_create(d->display, fd);
+    if (!c) {
+        return nullptr;
+    }
+    return getConnection(c);
+}
 
 }
 }
