@@ -18,6 +18,8 @@ You should have received a copy of the GNU Lesser General Public
 License along with this library.  If not, see <http://www.gnu.org/licenses/>.
 *********************************************************************/
 #include "shell_interface.h"
+#include "global_p.h"
+#include "resource_p.h"
 #include "display.h"
 #include "surface_interface.h"
 
@@ -32,73 +34,53 @@ namespace Server
 
 static const quint32 s_version = 1;
 
-class ShellInterface::Private
+class ShellInterface::Private : public Global::Private
 {
 public:
     Private(ShellInterface *q, Display *d);
-    void create();
 
-    Display *display;
-    wl_global *shell = nullptr;
     QList<ShellSurfaceInterface*> surfaces;
 
 private:
-    static void bind(wl_client *client, void *data, uint32_t version, uint32_t id);
     static void createSurfaceCallback(wl_client *client, wl_resource *resource, uint32_t id, wl_resource *surface);
-    void bind(wl_client *client, uint32_t version, uint32_t id);
-    void createSurface(wl_client *client, uint32_t version, uint32_t id, SurfaceInterface *surface);
+    void bind(wl_client *client, uint32_t version, uint32_t id) override;
+    void createSurface(wl_client *client, uint32_t version, uint32_t id, SurfaceInterface *surface, wl_resource *parentResource);
 
     ShellInterface *q;
     static const struct wl_shell_interface s_interface;
 };
 
 ShellInterface::Private::Private(ShellInterface *q, Display *d)
-    : display(d)
+    : Global::Private(d, &wl_shell_interface, s_version)
     , q(q)
 {
 }
 
-void ShellInterface::Private::create()
-{
-    Q_ASSERT(!shell);
-    shell = wl_global_create(*display, &wl_shell_interface, s_version, this, &bind);
-}
-
 const struct wl_shell_interface ShellInterface::Private::s_interface = {
     createSurfaceCallback
 };
 
 
 
-class ShellSurfaceInterface::Private
+class ShellSurfaceInterface::Private : public Resource::Private
 {
 public:
-    Private(ShellSurfaceInterface *q, ShellInterface *shell, SurfaceInterface *surface);
-    void create(wl_client *client, quint32 version, quint32 id);
+    Private(ShellSurfaceInterface *q, ShellInterface *shell, SurfaceInterface *surface, wl_resource *parentResource);
     void setFullscreen(bool fullscreen);
     void setToplevel(bool toplevel);
+    void setMaximized(bool maximized);
     void ping();
 
     SurfaceInterface *surface;
-    ShellInterface *shell;
-    wl_resource *shellSurface = nullptr;
-    wl_client *client = nullptr;
-    pid_t clientPid = 0;
-    uid_t clientUser = 0;
-    gid_t clientGroup = 0;
     QString title;
     QByteArray windowClass;
     QScopedPointer<QTimer> pingTimer;
     quint32 pingSerial = 0;
     bool fullscreen = false;
     bool toplevel = false;
+    bool maximized = false;
 
 private:
-    static Private *cast(wl_resource *r) {
-        return reinterpret_cast<Private*>(wl_resource_get_user_data(r));
-    }
-
-    static void unbind(wl_resource *r);
     // interface callbacks
     static void pongCallback(wl_client *client, wl_resource *resource, uint32_t serial);
     static void moveCallback(wl_client *client, wl_resource *resource, wl_resource *seat, uint32_t serial);
@@ -118,44 +100,24 @@ private:
     void setTitle(const QString &title);
     void setWindowClass(const QByteArray &windowClass);
     void pong(quint32 serial);
+    ShellSurfaceInterface *q_func() {
+        return reinterpret_cast<ShellSurfaceInterface *>(q);
+    }
 
-    ShellSurfaceInterface *q;
     static const struct wl_shell_surface_interface s_interface;
 };
 
 ShellInterface::ShellInterface(Display *display, QObject *parent)
-    : QObject(parent)
-    , d(new Private(this, display))
+    : Global(new Private(this, display), parent)
 {
 }
 
-ShellInterface::~ShellInterface()
-{
-    destroy();
-}
-
-void ShellInterface::create()
-{
-    d->create();
-}
-
-void ShellInterface::destroy()
-{
-    if (!d->shell) {
-        return;
-    }
-    wl_global_destroy(d->shell);
-    d->shell = nullptr;
-}
-
-void ShellInterface::Private::bind(wl_client *client, void *data, uint32_t version, uint32_t id)
-{
-    reinterpret_cast<ShellInterface::Private*>(data)->bind(client, version, id);
-}
+ShellInterface::~ShellInterface() = default;
 
 void ShellInterface::Private::bind(wl_client *client, uint32_t version, uint32_t id)
 {
-    wl_resource *shell = wl_resource_create(client, &wl_shell_interface, qMin(version, s_version), id);
+    auto c = display->getConnection(client);
+    wl_resource *shell = c->createResource(&wl_shell_interface, qMin(version, s_version), id);
     if (!shell) {
         wl_client_post_no_memory(client);
         return;
@@ -166,49 +128,38 @@ void ShellInterface::Private::bind(wl_client *client, uint32_t version, uint32_t
 void ShellInterface::Private::createSurfaceCallback(wl_client *client, wl_resource *resource, uint32_t id, wl_resource *surface)
 {
     auto s = reinterpret_cast<ShellInterface::Private*>(wl_resource_get_user_data(resource));
-    s->createSurface(client, wl_resource_get_version(resource), id, SurfaceInterface::get(surface));
+    s->createSurface(client, wl_resource_get_version(resource), id, SurfaceInterface::get(surface), resource);
 }
 
-void ShellInterface::Private::createSurface(wl_client *client, uint32_t version, uint32_t id, SurfaceInterface *surface)
+void ShellInterface::Private::createSurface(wl_client *client, uint32_t version, uint32_t id, SurfaceInterface *surface, wl_resource *parentResource)
 {
     auto it = std::find_if(surfaces.constBegin(), surfaces.constEnd(),
         [surface](ShellSurfaceInterface *s) {
             return surface == s->surface();
         }
     );
-    if (it != surfaces.constBegin()) {
-        wl_resource_post_error(surface->surface(), WL_DISPLAY_ERROR_INVALID_OBJECT, "ShellSurface already created");
+    if (it != surfaces.constEnd()) {
+        wl_resource_post_error(surface->resource(), WL_DISPLAY_ERROR_INVALID_OBJECT, "ShellSurface already created");
         return;
     }
-    ShellSurfaceInterface *shellSurface = new ShellSurfaceInterface(q, surface);
+    ShellSurfaceInterface *shellSurface = new ShellSurfaceInterface(q, surface, parentResource);
     surfaces << shellSurface;
     QObject::connect(shellSurface, &ShellSurfaceInterface::destroyed, q,
         [this, shellSurface] {
             surfaces.removeAll(shellSurface);
         }
     );
-    shellSurface->d->create(client, version, id);
+    shellSurface->d->create(display->getConnection(client), version, id);
     emit q->surfaceCreated(shellSurface);
 }
 
-bool ShellInterface::isValid() const
-{
-    return d->shell != nullptr;
-}
-
-Display *ShellInterface::display() const
-{
-    return d->display;
-}
-
 /*********************************
  * ShellSurfaceInterface
  *********************************/
-ShellSurfaceInterface::Private::Private(ShellSurfaceInterface *q, ShellInterface *shell, SurfaceInterface *surface)
-    : surface(surface)
-    , shell(shell)
+ShellSurfaceInterface::Private::Private(ShellSurfaceInterface *q, ShellInterface *shell, SurfaceInterface *surface, wl_resource *parentResource)
+    : Resource::Private(q, shell, parentResource, &wl_shell_surface_interface, &s_interface)
+    , surface(surface)
     , pingTimer(new QTimer)
-    , q(q)
 {
     pingTimer->setSingleShot(true);
     pingTimer->setInterval(1000);
@@ -227,17 +178,19 @@ const struct wl_shell_surface_interface ShellSurfaceInterface::Private::s_interf
     setClassCallback
 };
 
-ShellSurfaceInterface::ShellSurfaceInterface(ShellInterface *shell, SurfaceInterface *parent)
-    : QObject(parent)
-    , d(new Private(this, shell, parent))
+ShellSurfaceInterface::ShellSurfaceInterface(ShellInterface *shell, SurfaceInterface *parent, wl_resource *parentResource)
+    : Resource(new Private(this, shell, parent, parentResource), parent)
 {
+    Q_D();
     connect(d->pingTimer.data(), &QTimer::timeout, this, &ShellSurfaceInterface::pingTimeout);
     connect(this, &ShellSurfaceInterface::fullscreenChanged, this,
         [this] (bool fullscreen) {
             if (!fullscreen) {
                 return;
             }
+            Q_D();
             d->setToplevel(false);
+            d->setMaximized(false);
         }
     );
     connect(this, &ShellSurfaceInterface::toplevelChanged, this,
@@ -245,44 +198,29 @@ ShellSurfaceInterface::ShellSurfaceInterface(ShellInterface *shell, SurfaceInter
             if (!toplevel) {
                 return;
             }
+            Q_D();
             d->setFullscreen(false);
+            d->setMaximized(false);
+        }
+    );
+    connect(this, &ShellSurfaceInterface::maximizedChanged, this,
+        [this] (bool maximized) {
+            if (!maximized) {
+                return;
+            }
+            Q_D();
+            d->setFullscreen(false);
+            d->setToplevel(false);
         }
     );
 }
 
-ShellSurfaceInterface::~ShellSurfaceInterface()
-{
-    if (d->shellSurface) {
-        wl_resource_destroy(d->shellSurface);
-    }
-}
-
-void ShellSurfaceInterface::Private::create(wl_client *c, quint32 version, quint32 id)
-{
-    Q_ASSERT(!client);
-    Q_ASSERT(!shellSurface);
-    shellSurface = wl_resource_create(c, &wl_shell_surface_interface, version, id);
-    if (!shellSurface) {
-        wl_client_post_no_memory(c);
-        return;
-    }
-    client = c;
-    wl_client_get_credentials(client, &clientPid, &clientUser, &clientGroup);
-
-    wl_resource_set_implementation(shellSurface, &s_interface, this, unbind);
-}
-
-void ShellSurfaceInterface::Private::unbind(wl_resource *r)
-{
-    auto s = cast(r);
-    s->shellSurface = nullptr;
-    s->q->deleteLater();
-}
+ShellSurfaceInterface::~ShellSurfaceInterface() = default;
 
 void ShellSurfaceInterface::Private::pongCallback(wl_client *client, wl_resource *resource, uint32_t serial)
 {
-    auto s = cast(resource);
-    Q_ASSERT(client == s->client);
+    auto s = cast<Private>(resource);
+    Q_ASSERT(client == *s->client);
     s->pong(serial);
 }
 
@@ -290,12 +228,14 @@ void ShellSurfaceInterface::Private::pong(quint32 serial)
 {
     if (pingTimer->isActive() && serial == pingSerial) {
         pingTimer->stop();
+        Q_Q(ShellSurfaceInterface);
         emit q->pongReceived();
     }
 }
 
 void ShellSurfaceInterface::ping()
 {
+    Q_D();
     d->ping();
 }
 
@@ -304,35 +244,38 @@ void ShellSurfaceInterface::Private::ping()
     if (pingTimer->isActive()) {
         return;
     }
-    pingSerial = shell->display()->nextSerial();
-    wl_shell_surface_send_ping(shellSurface, pingSerial);
-    wl_client_flush(client);
+    pingSerial = global->display()->nextSerial();
+    wl_shell_surface_send_ping(resource, pingSerial);
+    client->flush();
     pingTimer->start();
 }
 
 void ShellSurfaceInterface::setPingTimeout(uint msec)
 {
+    Q_D();
     d->pingTimer->setInterval(msec);
 }
 
 bool ShellSurfaceInterface::isPinged() const
 {
+    Q_D();
     return d->pingTimer->isActive();
 }
 
 void ShellSurfaceInterface::requestSize(const QSize &size)
 {
+    Q_D();
     // TODO: what about the edges?
-    wl_shell_surface_send_configure(d->shellSurface, 0, size.width(), size.height());
-    wl_client_flush(d->client);
+    wl_shell_surface_send_configure(d->resource, 0, size.width(), size.height());
+    d->client->flush();
 }
 
 void ShellSurfaceInterface::Private::moveCallback(wl_client *client, wl_resource *resource, wl_resource *seat, uint32_t serial)
 {
     Q_UNUSED(seat)
     Q_UNUSED(serial)
-    auto s = cast(resource);
-    Q_ASSERT(client == s->client);
+    auto s = cast<Private>(resource);
+    Q_ASSERT(client == *s->client);
     // TODO: implement
 }
 
@@ -341,15 +284,15 @@ void ShellSurfaceInterface::Private::resizeCallback(wl_client *client, wl_resour
     Q_UNUSED(seat)
     Q_UNUSED(serial)
     Q_UNUSED(edges)
-    auto s = cast(resource);
-    Q_ASSERT(client == s->client);
+    auto s = cast<Private>(resource);
+    Q_ASSERT(client == *s->client);
     // TODO: implement
 }
 
 void ShellSurfaceInterface::Private::setToplevelCallback(wl_client *client, wl_resource *resource)
 {
-    auto s = cast(resource);
-    Q_ASSERT(client == s->client);
+    auto s = cast<Private>(resource);
+    Q_ASSERT(client == *s->client);
     s->setToplevel(true);
 }
 
@@ -359,6 +302,7 @@ void ShellSurfaceInterface::Private::setToplevel(bool t)
         return;
     }
     toplevel = t;
+    Q_Q(ShellSurfaceInterface);
     emit q->toplevelChanged(toplevel);
 }
 
@@ -369,8 +313,8 @@ void ShellSurfaceInterface::Private::setTransientCallback(wl_client *client, wl_
     Q_UNUSED(x)
     Q_UNUSED(y)
     Q_UNUSED(flags)
-    auto s = cast(resource);
-    Q_ASSERT(client == s->client);
+    auto s = cast<Private>(resource);
+    Q_ASSERT(client == *s->client);
     // TODO: implement
 }
 
@@ -380,8 +324,8 @@ void ShellSurfaceInterface::Private::setFullscreenCallback(wl_client *client, wl
     Q_UNUSED(method)
     Q_UNUSED(framerate)
     Q_UNUSED(output)
-    auto s = cast(resource);
-    Q_ASSERT(client == s->client);
+    auto s = cast<Private>(resource);
+    Q_ASSERT(client == *s->client);
     // TODO: add method, framerate and output
     s->setFullscreen(true);
 }
@@ -392,6 +336,7 @@ void ShellSurfaceInterface::Private::setFullscreen(bool f)
         return;
     }
     fullscreen = f;
+    Q_Q(ShellSurfaceInterface);
     emit q->fullscreenChanged(fullscreen);
 }
 
@@ -404,23 +349,33 @@ void ShellSurfaceInterface::Private::setPopupCalback(wl_client *client, wl_resou
     Q_UNUSED(x)
     Q_UNUSED(y)
     Q_UNUSED(flags)
-    auto s = cast(resource);
-    Q_ASSERT(client == s->client);
+    auto s = cast<Private>(resource);
+    Q_ASSERT(client == *s->client);
     // TODO: implement
 }
 
 void ShellSurfaceInterface::Private::setMaximizedCallback(wl_client *client, wl_resource *resource, wl_resource *output)
 {
     Q_UNUSED(output)
-    auto s = cast(resource);
-    Q_ASSERT(client == s->client);
-    // TODO: implement
+    auto s = cast<Private>(resource);
+    Q_ASSERT(client == *s->client);
+    s->setMaximized(true);
+}
+
+void ShellSurfaceInterface::Private::setMaximized(bool set)
+{
+    if (maximized == set) {
+        return;
+    }
+    maximized = set;
+    Q_Q(ShellSurfaceInterface);
+    emit q->maximizedChanged(maximized);
 }
 
 void ShellSurfaceInterface::Private::setTitleCallback(wl_client *client, wl_resource *resource, const char *title)
 {
-    auto s = cast(resource);
-    Q_ASSERT(client == s->client);
+    auto s = cast<Private>(resource);
+    Q_ASSERT(client == *s->client);
     s->setTitle(QString::fromUtf8(title));
 }
 
@@ -430,13 +385,14 @@ void ShellSurfaceInterface::Private::setTitle(const QString &t)
         return;
     }
     title = t;
+    Q_Q(ShellSurfaceInterface);
     emit q->titleChanged(title);
 }
 
 void ShellSurfaceInterface::Private::setClassCallback(wl_client *client, wl_resource *resource, const char *class_)
 {
-    auto s = cast(resource);
-    Q_ASSERT(client == s->client);
+    auto s = cast<Private>(resource);
+    Q_ASSERT(client == *s->client);
     s->setWindowClass(QByteArray(class_));
 }
 
@@ -446,47 +402,48 @@ void ShellSurfaceInterface::Private::setWindowClass(const QByteArray &wc)
         return;
     }
     windowClass = wc;
+    Q_Q(ShellSurfaceInterface);
     emit q->windowClassChanged(windowClass);
 }
 
 SurfaceInterface *ShellSurfaceInterface::surface() const {
+    Q_D();
     return d->surface;
 }
 
 ShellInterface *ShellSurfaceInterface::shell() const {
-    return d->shell;
-}
-
-wl_resource *ShellSurfaceInterface::shellSurface() const {
-    return d->shellSurface;
+    Q_D();
+    return reinterpret_cast<ShellInterface*>(d->global);
 }
 
 QString ShellSurfaceInterface::title() const {
+    Q_D();
     return d->title;
 }
 
 QByteArray ShellSurfaceInterface::windowClass() const {
+    Q_D();
     return d->windowClass;
 }
 
 bool ShellSurfaceInterface::isFullscreen() const {
+    Q_D();
     return d->fullscreen;
 }
 
 bool ShellSurfaceInterface::isToplevel() const {
+    Q_D();
     return d->toplevel;
 }
 
-pid_t ShellSurfaceInterface::clientPid() const {
-    return d->clientPid;
+bool ShellSurfaceInterface::isMaximized() const {
+    Q_D();
+    return d->maximized;
 }
 
-uid_t ShellSurfaceInterface::clientUser() const {
-    return d->clientUser;
-}
-
-gid_t ShellSurfaceInterface::clientGroup() const {
-    return d->clientGroup;
+ShellSurfaceInterface::Private *ShellSurfaceInterface::d_func() const
+{
+    return reinterpret_cast<ShellSurfaceInterface::Private*>(d.data());
 }
 
 }
